/*
 * Hormann, Leydold, Derflinger
 * Automatic Nonuniform Random Variate Generation
 * Springer, Heidelberg, New York, 2003
 * http://statistik.wu-wien.ac.at/arvag/monograph
 * 
 * this file: algo.h
 *
 * header-file with the function prototypes for all implemented algorithms
 *
 */


double generator02_02(double (*pdf)(double), double bl, double br, double m);
/*
 * Algorithm 2.2 (rejection from uniform hat) p.17 
 *
 * pdf... function pointer to the density
 * bl ... left border of domain
 * br ... right border of domain
 * m  ... upper bound for the density.
 */

double generator02_09(double (*pdf)(double), double mu, double um, double up, double vp);
/*
 * Algorithm 2.9 (Ratio-of-uniforms), p. 35 
 *
 * pdf... function pointer to the density
 * mu ... constant (should be close to the mode of the distribution)
 * um ... u- left border of (minimal) bounding rectangle 
 * up ... u+ right border of (minimal) bounding rectangle 
 * vp ... upper border of (minimal) bounding rectangle 
 */

double generator02_11(void);
/*
 * Algorithm 2.11 (Box-Muller), p. 38  
 */

int generator03_01(double (*pmf)(int));
/*
 * Algorithm 3.1 (Sequential-Search), p. 44  
 * generates discrete variates
 *
 * pmf... function pointer to probability (mass) function for k >= 0
 */

int setup03_02(double (*pmf)(int), int L, int C);
/*
 * Algorithm 3.2 (Indexed-Search), p. 46  
 * generates discrete variates
 *
 * pmf... function pointer to probability (mass) function for k >= 0
 * L ... non zero probabilities assumed for (0, 1, 2, ..., L-1)
 * C ... length of guide table
 *
 */

int generator03_02();
/* is called after the setup to generate random variates */

     int free03_02();
/* to free the allocated memory*/


int setup03_03(double (*pmf)(int), int L);
/*
 * Algorithm 3.4 (Alias setup for Algorithm 3.3), p. 48
 *
 * pmf... function pointer to probability (mass) function for k >= 0
 * L ... non zero probabilities assumed for (0, 1, 2, ..., L-1)
 *
 */
     int generator03_03();
/*
 * Algorithm 3.3 (Alias sample), p. 48
 *
 */
     int free03_03();
/* frees the allocated memory */


int setup04_01(double (*pdf)(double), double (*dpdf)(double), 
                 double bl, double br, 
		  int N, double constructionp[]);
/*
 * Algorithm 4.1 (TDR Transformed-Density-Rejection), p. 60 
 *
 * pdf... function pointer to the density
 * dpdf... function pointer to the derivative of the density
 * bl ... left border of domain
 * br ... right border of domain
 * N  ... number of construction points
 * constructionp[] ... array holding the construction points
 *
 */
     double generator04_01();
/* is called after the setup to generate random variates */
     int free04_01();
/* frees the allocated memory */

int setup04_02(double (*pdf)(double), double (*dpdf)(double), 
                 double bl, double br, 
	       int N, double constructionp[]);
/*
 * Algorithm 4.2 (TDR proprtional squeeze), p. 91   
 *
 * setup ... 1 = new setup; 0 = sample; -1 = free tables
 * pdf... function pointer to the density
 * dpdf... function pointer to the derivative of the density
 * bl ... left border of domain
 * br ... right border of domain
 * N  ... number of construction points
 * constructionp[] ... array holding the construction points
 *
 */
     double generator04_02();
/* is called after the setup to generate random variates */
     int free04_02();
/* frees the allocated memory */


int setup04_03(  double (*pdf)(double), double (*dpdf)(double), 
                 double bl, double br, 
                 int N, double constructionp[]);
/*
 * Algorithm 4.3 (TDR immediate acceptance), p. 92  
 *
 * setup ... 1 = new setup; 0 = sample; -1 = free tables
 * pdf... function pointer to the density
 * dpdf... function pointer to the derivative of the density
 * bl ... left border of domain
 * br ... right border of domain
 * N  ... number of construction points
 * constructionp[] ... array holding the construction points
 *
 */

     double generator04_03();
/* is called after the setup to generate random variates */

     int free04_03();
/* frees the allocated memory */

int setup04_04(  double (*pdf)(double),  
                 double bl, double m, double br);
/*
 * Algorithm 4.4 (TDR-3-Points (for c = -1/2)), p. 94  
 *
 * setup ... 1 = new setup; 0 = sample; -1 = free tables
 * pdf... function pointer to the density
 * bl ... left border of domain; bl < -1.e99 is treated as -infinity
 * m  ... mode of the density
 * br ... right border of domain; br > 1.e99 is treated as +infinity
 *
 */
     double generator04_04();
/* is called after the setup to generate random variates */
int setup04_05(  double (*pdf)(double), double (*dpdf)(double),  
                 int N, double constructionp[]);
/*
 * Algorithm 4.5 (TDR-T-convex), p. 99  
 *
 * setup ... 1 = new setup; 0 = sample; -1 = free tables
 * pdf... function pointer to the density
 * dpdf... function pointer to the derivative of the density
 * N  ... number of construction points
 * constructionp[] ... array holding the construction points
 * constructionp[0] is the left, constructionp[N-1] the right border of the domain
 *
 */
     double generator04_05();
/* is called after the setup to generate random variates */
     int free04_05();
/* frees the allocated memory */ 

int setup04_07(double (*pdf)(double), 
               double (*dpdf)(double),  double (*ddpdf)(double),  
	       int N, double constructionp[]);
/*
 * Algorithm 4.7 (TDR-c-optimal), p. 102  
 *
 * pdf... function pointer to the density
 * dpdf... function pointer to the derivative of the density
 * ddpdf...function pointer to the second derivative of the density
 * N  ... number of construction points
 * constructionp[] ... array holding the construction points
 * constructionp[0] is the left, constructionp[N-1] the right border of the domain
 *
 */
double generator04_07();
/* is called after the setup to generator reandom variates */
int free04_07();
/* frees the allocated memory */
 

int setup05_01( double (*pdf)(double), 
                int N, double constructionp[]);
/*
 * Algorithm 5.1 (Ahrens, basic method), p. 115  
 *
 * pdf... function pointer to the density
 * N  ... is the number of construction points (in the book this is N+1)
 * constructionp[] ... array holding the construction points
 * constructionp[0] is the left, constructionp[N-1] the right border of the domain
 *
 */

     double generator05_01();
/* is called after the setup to generator reandom variates */

     int free05_01();
/* frees the allocated memory */


int setup05_02( double (*pdf)(double), 
                int N, double constructionp[]);
/*
 * Algorithm 5.2 (Ahrens, immediate acceptance), p. 116  
 *
 * pdf... function pointer to the density
 * N  ... is the number of construction points (in the book this is N+1)
 * constructionp[] ... array holding the construction points
 * constructionp[0] is the left, constructionp[N-1] the right border of the domain
 *
 */

double generator05_02();
/* is called after the setup to generator reandom variates */

int free05_02();
/* frees the allocated memory */


int setup05_03( double (*pdf)(double), double Af,
                int NT, double bl, double br);
/*
 * Algorithm 5.3 (Ahrens, equal area), p. 122  
 *
 * pdf... function pointer to the density
 * Af ... approximate are below the pdf
 * NT  ... approximate number of construction points
 * bl ... left border of the domain
 * br ... right border of the domain
 *
 */

double generator05_03();
/* is called after the setup to generator reandom variates */

int free05_03();
/* frees the allocated memory */


int setup06_01( double (*pdf)(double), double Afb);
/*
 * Algorithm 6.1 (Monotone-01, simple rej. for monotone pdf on (0,1), p. 128)
 *
 * pdf ... function pointer to the density, which must be monotne on (0,1)
 * Afb ... upper bound for the area below the (quasi-)density 
 *
 */

double generator06_01();
/* is called after the setup to generate random variates */


int setup06_02( double (*pdf)(double), double Afb, double bl, double br);
/*
 * Algorithm 6.2 (Monotone-bounded, p. 128)
 *
 * pdf ... function pointer to the density, which must be monotne on (0,1)
 * Afb ... upper bound for the area below the (quasi-)density 
 * bl  ... left border of the domain
 * br  ... right border of the domain
 *
 */

double generator06_02();
/* is called after the setup to generate random variates */


int setup06_03( double (*pdf)(double), double ri, double muri);
/*
 * Algorithm 6.3 (Monotone-unbounded, p. 130)
 *
 * pdf ... function pointer to the density, which must be monotone on (0,infinity)
 *         (the pdf must be a real density, i.e. Af = 1.)
 * ri ... the r-th moment of the distribution has a known upper bound
 * muri ... upper bound for teh r-th moment
 */
double generator06_03();
/* is called after the setup to generate random variates */

int setup06_04( double (*pdf)(double), double Af, double m, double Fm);
/*
 * Algorithm 6.4 (Simple Ratio-of-Uniforms method, p. 138)
 *
 * pdf ... function pointer to the quasi density
 * Af ... area below the quasi density
 * m ... mode of the density
 * Fm ... CDF value at the mode m; if unknown -1. must be used
 */

double generator06_04();
/* is called after the setup to generate random variates */



int setup06_05( double (*pdf)(double), double Af, double mi);
/*
 * Algorithm 6.5 (Simple Ratio-of-Uniforms mirror principle, p. 139)
 *
 * pdf ... function pointer to the quasi density
 * Af ... area below the quasi density
 * mi ... mode of the density
 */

double generator06_05();
/* is called after the setup to generate random variates */


int setup06_06( double (*pdf)(double), double Af, double mi, double Fm);
/*
 * Algorithm 6.6 (Simple Setup rejection, p. 141)
 *
 * pdf ... function pointer to the quasi density
 * Af ... area below the quasi density
 * mi ... mode of the density
 * Fm ... CDF value at the mode m; if unknown -1. must be used
 */

double generator06_06();
/* is called after the setup to generate random variates */


int setup06_07( double (*pdf)(double), double Af, double mi, double Fm, double ri);
/*
 * Algorithm 6.7 (Generalized Simple RoU, p. 147)
 *
 * pdf ... function pointer to the quasi density
 * Af ... area below the quasi density
 * mi ... mode of the density
 * Fm ... CDF value at the mode m; if unknown -1. must be used
 * ri ... parameter of the method: we must have r > 1 and
 *       for the c of c-concavity we must have r >= -c/(1+c)
 */
double generator06_07();
/* is called after the setup to generate random variates */


int setup06_08( double (*pdf)(double), double Af, double mi, double Fm, double ri);
/*
 * Algorithm 6.8 (Heavy Tailed Simple RoU, p. 151)
 *
 * pdf ... function pointer to the quasi density; must be heavy-tailed T_c conave!!
           with c <= -1/2
 * Af ... area below the quasi density
 * mi ... mode of the density
 * Fm ... CDF value at the mode m; if unknown -1. must be used
 * ri ... parameter of the distribution: we must have r >= 1 and
 *       for the c of heavy tailed T-c concave  we must take r = -c/(1+c)
 */

double generator06_08();
/* is called after the setup to generate random variates */


int setup07_01( double (*cdf)(double), double (*pdf)(double),
                double eu, int nstart, int nmax, double bl, double br);
/*
 * Algorithm 7.1 (Fast-Table-Inversion, p. 163)
 *
 * cdf ... function pointer to the cumulative distribution function
 * pdf ... function pointer to the density
 * eu ... maxmial allowed error in u-direction
 * nstart ... number of starting intervals
 * nmax ... maximal number of intervals possible
 * bl ... approximate left border of domain, pdf(bl)>0 is required
 * br ... approximate right border of domain, pdf(br)>0 is required
 */

     double generator07_01();
/* is called after the setup to generate random variates */

     int free07_01();
/* frees the allocated memory */


double generator09_01(double (*ich)(double));
/*
 * Algorithm 9.1 (Hazard Rate Inversion), p. 195 
 *
 * ich... function pointer to the inverse cumulative hazard rate
 */

double generator09_03(double (*h)(double),double (*g)(double),
                      double (*icg)(double));
/*
 * Algorithm 9.3 (Hazard Rate Thinning), p. 196 
 *
 * h   ... function pointer to the hazard rate h(x)
 * g   ... function pointer to the dominatig hazard rate g(x)
 * icg ... function pointer to the inverse cumulative hazard rate of g(x)
 */

double generator09_04(double (*h)(double), double M);
/*
 * Algorithm 9.4 (Hazard Rate Constant Thinning), p. 197 
 *
 * h   ... function pointer to the bounded hazard rate h(x)
 * M   ... upper bound of the hazard rate h(x)
 */

double generator09_05(double (*h)(double));
/*
 * Algorithm 9.5 (Decreasing Hazard Rate dynamic Thinning), p. 198 
 *
 * h   ... function pointer to the bounded, decreasing hazard rate h(x)
 */

double generator09_06(double (*h)(double), double p0);
/*
 * Algorithm 9.6 (Increasing Hazard Rate dynamic composition) p.200 
 *
 * h   ... function pointer to the bounded, decreasing hazard rate h(x)
 * p0  ... design parameter of the algorithm, can be chosen close to the expectation
 *         of the respective distribution
 */

int setup10_01(double (*pmf)(int), int C);
/*
 * Algorithm 10.1 (Indexed-Search-Tail), p. 217
 * generates discrete variates
 *
 * pmf... function pointer to probability (mass) function for k >= 0
 * C ... length of guide table
 *
 */
     int generator10_01();
/* is called after the setup to generate random variates */
     int free10_01();
/* to free the allocated memory*/

int setup10_02(double (*pmf)(int), int ili, int iri, int C);
/*
 * Algorithm 10.2 (Indexed-Search-2-Tails), p. 220
 * generates discrete variates
 *
 * pmf... function pointer to probability (mass) function for k >= 0
 * ili iri ... (possibly short) interval that contains all modes of the pmf
 * C ... length of guide table
 *
 */
     int generator10_02();
/* is called after the setup to generate random variates */
     int free10_02();
/* to free the allocated memory*/

/* below is the structur for the parameters of the discrete distributions 
 * used in demo_algo10_04.c*/
typedef struct { double d1, d2, d3, d4;
                 long i1,i2,i3;   
	       }PARAM;

/* below is the structur of all constants of the generator 
 * used in algo10_04.c*/
typedef struct{
       double vt, vc, vcr, xsq[2], y[2], ys[2], ac[2], pm, Hat[2];
       long int m, n[2], x[2], sign[2],s[2];
       double (*hh)(PARAM *param,long x);
       PARAM par;
     }TOTAL;

TOTAL *setup10_04(double (*hh)(PARAM *param,long x),PARAM *param,long m,long la,long ra,double pma);
/*
 * Algorithm 10.4 (Automatic-Rejection-Inversion), p. 226
 *
 * the functioner returns a pointer to teh structure holding all necessary constants
 *
 * INPUTS:
 * hh....pointer to pmf, 
 *  param .... paramaters of pmf, 
 *  pma ... < 0 means pmf has sum 1; 
 *      pma>0 is the approximate mode probability as the sum of the pmf<>1
 *  m,la,ra...mode, smallest, largest number with probability>0
 *
 */

int generator10_04(TOTAL *t);
     /* calls the genrator */ 

int free10_04(TOTAL *t);
     /* frees the allocated memory */


int setup10_05(double (*pmf)(int), int mi, double s2i);
/*
 * Algorithm 10.5 (Discrete-Unimodal-2nd Moment), p. 229
 * generates discrete variates
 *
 * pmf... function pointer to unimodal probability (mass) function for k >= 0
 *        it need not sum up to 1.
 * mi ... mode of the pmf
 * s2i ... upper bound for teh second moment around 0 (i.e. Var + (m-mu)^2)
 *
 */
     int generator10_05();
/* is called after the setup to generate random variates */


int setup10_06(double (*pmf)(int), int mi, double S, double Fm1);
/*
 * Algorithm 10.6 (Discrete-Simple-RoU), p. 231
 * generates discrete variates
 *
 * pmf... function pointer to unimodal probability (mass) function for k >= 0
 *        it need not sum up to 1.
 * mi ... mode of the pmf
 * S ... sum of the probabilities of the pmf
 * Fm1 ... CDF value at m-1; if unknown -1. must be used
 *
 */
     int generator10_06();
/* is called after the setup to generate random variates */



int setup10_07(double (*pmf)(int), int mi, double S, double Fm1);
/*
 * Algorithm 10.7 (Discrete-Simple-RoU 2), p. 231
 * generates discrete variates
 *
 * pmf... function pointer to unimodal probability (mass) function for k >= 0
 *        it need not sum up to 1.
 * mi ... mode of the pmf
 * S ... sum of the probabilities of the pmf
 * Fm1 ... CDF value at m-1; if unknown -1. must be used
 *
 */
int generator10_07();
/* is called after the setup to generate random variates */


int generator10_09(double (*h)(int));
/*
 * Algorithm 10.9 (Discrete-Hazard-Rate-Inversion), p. 235
 * generates discrete variates
 *
 * h... function pointer to the hazard rate of a discrete distribution
 *
 */


int setup11_03(int d, double *S);
/*
 * Algorithm 11.3 (Multinormal-Cholesky, p. 250)
 *
 * d ... dimension (length) of the generated vectors
 * S ... pointer to the array that holds the consecutive columns 
 *       of the variance-covariance matrix
 */


void generator11_03(double x[]);
/* 
 * is called after the setup to generate a random vector 
 * x ... is the array to hold teh generated vector     
 */


int free11_03();
/* frees the allocated memory */

int algo11_04(int d, double *S, double *L);
/*
 * Algorithm 11.4 (Cholesky-Decomopsition, p. 251)
 *
 * d ... dimension (length) of the generated vectors
 * S ... pointer to the array that holds the consecutive columns 
 *       of the variance-covariance matrix
 * L ... pointer to the array that holds (as result of the algorithm)
 *       the consecutive rows of the Cholesky-factor
 * 
 * Output: 1 is returned if it was possible to calculate the Cholesky factor.
 *         0 if the matrix was not positive definite.
 */

void generator11_06(int d, double x[]);
/* 
 * Algorithm 11.6 (Uniform-Sphere, p. 253)
 * generates a random vector on the sphere S^(d-1)
 * d ... length of the generated vectors
 * x ... is the array to hold the generated vector     
 */

void generator11_07(int d, double x[]);
/* 
 * Algorithm 11.7 (Uniform-Ball-Rejection, p. 254)
 * generates a random vector in the ball B^d
 * d ... length of the generated vectors
 * x ... is the array to hold the generated vector     
 */

void generator11_08(int d, double x[]);
/* 
 * Algorithm 11.8 (Uniform-Ball, p. 255)
 * generates a random vector in the ball B^d
 * d ... length of the generated vectors
 * x ... is the array to hold the generated vector     
 */

void generator11_09(int d, double s[]);
/* 
 * Algorithm 11.9 (Uniform-Spacings, p. 257)
 * generates uniform spacings (S_1, S_2, ... S_d+1)
 * d+1 ... length of the generated vectors
 * s ... is the array to hold the generated vector     
 */

void generator11_10(int d, double v[],double x[]);
/* 
 * Algorithm 11.10 (Uniform-Simplex, p. 257)
 * generates a vector uniform distributed over the simplex with vertices v_i
 * d ... length of the generated vectors (must be smaller than 30)
 * v ... is the array that contains the (d+1) vertices stored in a
 *       sequential way.
 * x ... is the array to hold the generated vector     
 */


double generator12_03(double x[], int n);
/*
 * Algorithm 12.3 (Resampling-ELK), p. 311 
 *
 * x[] ... array holding the ORDERED!! sample in x[1],x[2],...,x[n]
 * n ... samplesize
 */


int setup12_05(double m1,double m2,double m3,double m4);
/*
 * Algorithm 12.5 (4 Moments discrete), p. 325 
 *
 * m1, m2, m3, m4 .... array holding the first four moments 
 */
     double generator12_05();
/* generates a random variate with the desired moments */


int setup12_06(double m1,double m2,double m3,double m4);
/*
 * Algorithm 12.6 (4 Moments continuous), p. 326 
 * with mixture distribution s*Z+1
 *
 * m1, m2, m3, m4 .... array holding the first four moments 
 */
double generator12_06();
/* generates a continuous random variate with the desired moments */


void generator12_07(double rho, double res[2]);
/*
 * Algorithm 12.7 (Zero-Maximal-Bivariate), p. 331 
 *
 * rho .... Correlation 
 * res[2] ... to hold the result
 *
 */


void generator12_08(double rho, double y[2]);
/*
 * Algorithm 12.8 (NORTU-Bivariate), p. 334 
 *
 * rho .... Correlation 
 * y[2] ... to hold the result
 *
 */

int setup12_09(double rho);
/*
 * Algorithm 12.9 (Chessboard-Bivariate), p. 335 
 *
 * rho .... Correlation 
 *
 */
void generator12_09(double res[2]);
/* generates random pairs with the desired correlation */


double generator14_06( double (*pdf)(double), double bl, double br, double s);
/*
 * Algorithm 14.6 (Perfect Random Walk Metropolis Sampler, p. 382)
 *
 * pdf ... function pointer to the density, which must be monotone
 * bl  ... left border of the domain
 * br  ... right border of the domain
 * s  ... random walk proposal is U(-s,s)
 *
 */
